import numpy as np
import math
import time
import sys
if sys.version_info.major == 2:
    import Tkinter as tk
else:
    import tkinter as tk
np.set_printoptions(suppress=True)
np.random.seed(1)

G = 9.81
ROLL_RATE = 40                                  # roll rate

class AirCombatEnv():
    def __init__(self):
        self.AREA = 3000                        # area

        # visualization
        self.SCALE = 0.1  # scale
        self.SCOPE = self.AREA * self.SCALE     # visualization

        # aircraft parameters
        self.ac_speed = 200                     # speed，m/s
        self.ac_bank_angle_b_max = 75           # φ_blue_max = 75
        self.ac_bank_angle_r_max = 75           # φ_red_max = 75

        # training parameters
        self.t = 0.5                            # training time step
        self.td = self.t / 5                    # simulation time step
        self.max_step = 1000                    # maximum training time step

        # advantage & disadvantage
        self.AA_range = 60              # deviation angle
        self.ATA_range = 30             # aspect angle
        self.Dis_max = 500              # d_max
        self.Dis_min = 100              # d_min
        self.adv_count = 0              # accumulate advantage time

        # initialization：0 random，1 advantage，2 disadvantage，3 head-on，4 neutral
        self.init_scen = 0

        # action
        self.action_space = ['l','s','r']       #left, maintain, right
        self.n_actions = len(self.action_space)
        self.n_features = 3
        self.disadvantage = 0
        self.dis_count = 0

    def reset_selfPlay(self):
        self.disadvantage = 0
        self.reward_b = 0
        self.reward_r = 0
        self.done = False
        self.success = 0
        self.acts = [[],[]]
        self.advs = []
        self.ATA_b = self.AA_b = 100
        self.ATA_r = self.AA_r = 100
        self.adv_count = 0

        # red_aircraft initialization
        self.ac_pos_r = np.array([0.0,0.0])         # coordinate
        self.ac_heading_r = np.array([0])           # heading angle
        self.ac_bank_angle_r = 0                    # bank angle

        # blue_aircraft initialization
        if self.init_scen == 0:         # random
            self.ac_pos_b = np.append(np.random.uniform(-500,500,size=1),np.random.uniform(-250,250,size=1))
            self.ac_heading_b = np.random.uniform(0,360,size=1)
        elif self.init_scen == 1:       # advantage
            self.ac_pos_b = np.append(np.random.uniform(-100,-500,size=1),np.random.uniform(-50,50,size=1))
            self.ac_heading_b = np.random.uniform(0,30,size=1)
        elif self.init_scen == 2:       # disadvantage
            self.ac_pos_b = np.append(np.random.uniform(100,500,size=1),np.random.uniform(-50,50,size=1))
            self.ac_heading_b = np.random.uniform(0,30,size=1)
        elif self.init_scen == 3:       # head-on
            self.ac_pos_b = np.append(np.random.uniform(100, 500, size=1), np.random.uniform(-50, 50, size=1))
            self.ac_heading_b = np.random.uniform(150, 180, size=1)
        else:                           # neutral
            #self.ac_pos_b = np.append(np.random.uniform(-50,50,size=1),np.random.uniform(-250,250,size=1))
            self.ac_pos_b = np.array([0.0, 250])
            self.ac_heading_b = np.array([0])

        self.ac_bank_angle_b = 0

        self.step_left = np.array([int(self.max_step)])

        dis = math.sqrt((self.ac_pos_r[0] - self.ac_pos_b[0]) * (self.ac_pos_r[0] - self.ac_pos_b[0])
                        + (self.ac_pos_r[1] - self.ac_pos_b[1]) * (self.ac_pos_r[1] - self.ac_pos_b[1]))

        # advantage
        heading_r = self.ac_heading_r
        heading_b = self.ac_heading_b

        self.ATA_b, self.AA_b = self.getAngle(self.ac_pos_r, self.ac_pos_b, heading_r, heading_b)
        self.ATA_r, self.AA_r = self.getAngle(self.ac_pos_b, self.ac_pos_r, heading_b, heading_r)

        if (dis < self.Dis_max) and (dis > self.Dis_min) \
                and (abs(self.AA_b) < self.AA_range) and (abs(self.ATA_b) < self.ATA_range):
            if self.adv_count >= 0:
                self.adv_count += 1
            else:
                self.adv_count = 1
        elif(dis < self.Dis_max) and (dis > self.Dis_min) \
                and (abs(self.AA_r) < self.AA_range) and (abs(self.ATA_r) < self.ATA_range):
            if self.adv_count <= 0:
                self.adv_count -= 1
            else:
                self.adv_count = -1
        else:
            self.adv_count = 0
        self.advs.append(self.adv_count)

        # reward shaping
        RA_b = 1 - ((1 - math.fabs(self.ATA_b) / 180) + (1 - math.fabs(self.AA_b) / 180))
        RA_r = 1 - ((1 - math.fabs(self.ATA_r) / 180) + (1 - math.fabs(self.AA_r) / 180))
        RD = math.exp(-(math.fabs(dis - ((self.Dis_max + self.Dis_min) / 2)) / 180 * 0.1))
        Rbl_b = RA_b * RD
        Rbl_r = RA_r * RD
        self.dis_count = Rbl_r
        self.fai_b = -0.01 * Rbl_b
        self.fai_r = -0.01 * Rbl_r

        #state
        s_b = np.concatenate(((self.ac_pos_b-self.ac_pos_r)/self.AREA,
                            [self.ac_heading_b/180, self.ac_heading_r/180,
                             self.ac_bank_angle_b/80,self.adv_count/10]))
        s_r = np.concatenate(((self.ac_pos_r-self.ac_pos_b)/self.AREA,
                            [self.ac_heading_r/180, self.ac_heading_b/180,
                             self.ac_bank_angle_r/80, 0-self.adv_count/10]))
        return s_b, s_r

    def step_selfPlay(self, action_b, action_r):
        # save action sequence
        self.acts[0].append(action_b)
        self.acts[1].append(action_r)

        for i in range(5):
            # blue aircraft dynamic
            self.ac_bank_angle_b = self.ac_bank_angle_b + (action_b - 1) * ROLL_RATE * self.td
            self.ac_bank_angle_b = max(self.ac_bank_angle_b,-self.ac_bank_angle_b_max)
            self.ac_bank_angle_b = min(self.ac_bank_angle_b, self.ac_bank_angle_b_max)
            self.turn_rate = (G / self.ac_speed) * math.tan(self.ac_bank_angle_b * math.pi / 180) * 180 / math.pi
            self.ac_heading_b = self.ac_heading_b - self.turn_rate * self.td
            if self.ac_heading_b > 360:
                self.ac_heading_b -= 360
            elif self.ac_heading_b < 0:
                self.ac_heading_b += 360
            self.ac_pos_b[0] = self.ac_pos_b[0] + self.ac_speed * self.td * math.cos(self.ac_heading_b * math.pi / 180)
            self.ac_pos_b[1] = self.ac_pos_b[1] + self.ac_speed * self.td * math.sin(self.ac_heading_b * math.pi / 180)

            # red aircraft dynamic
            self.ac_bank_angle_r = self.ac_bank_angle_r + (action_r - 1) * ROLL_RATE * self.td
            self.ac_bank_angle_r = max(self.ac_bank_angle_r, -self.ac_bank_angle_r_max)
            self.ac_bank_angle_r = min(self.ac_bank_angle_r, self.ac_bank_angle_r_max)
            self.turn_rate = (G / self.ac_speed) * math.tan(self.ac_bank_angle_r * math.pi / 180) * 180 / math.pi
            self.ac_heading_r = self.ac_heading_r - self.turn_rate * self.td
            if self.ac_heading_r > 360:
                self.ac_heading_r -= 360
            elif self.ac_heading_r < 0:
                self.ac_heading_r += 360
            self.ac_pos_r[0] = self.ac_pos_r[0] + self.ac_speed * self.td * math.cos(self.ac_heading_r * math.pi / 180)
            self.ac_pos_r[1] = self.ac_pos_r[1] + self.ac_speed * self.td * math.sin(self.ac_heading_r * math.pi / 180)
            if self.dis_count < 0:
                self.disadvantage += 0.1

        self.step_left -= 1

        # next state
        s_b = np.concatenate(((self.ac_pos_b-self.ac_pos_r)/self.AREA,
                            [self.ac_heading_b/180, self.ac_heading_r/180,
                             self.ac_bank_angle_b/80,self.adv_count/10]))
        s_r = np.concatenate(((self.ac_pos_r-self.ac_pos_b)/self.AREA,
                            [self.ac_heading_r/180, self.ac_heading_b/180,
                             self.ac_bank_angle_r/80, 0-self.adv_count/10]))

        dis = math.sqrt((self.ac_pos_r[0] - self.ac_pos_b[0]) * (self.ac_pos_r[0] - self.ac_pos_b[0])
                        + (self.ac_pos_r[1] - self.ac_pos_b[1]) * (self.ac_pos_r[1] - self.ac_pos_b[1]))

        #advantage
        heading_r = self.ac_heading_r
        heading_b = self.ac_heading_b

        self.ATA_b, self.AA_b = self.getAngle(self.ac_pos_r, self.ac_pos_b, heading_r, heading_b)
        self.ATA_r, self.AA_r = self.getAngle(self.ac_pos_b, self.ac_pos_r, heading_b, heading_r)

        if (dis < self.Dis_max) and (dis > self.Dis_min) \
                and (abs(self.AA_b) < self.AA_range) and (abs(self.ATA_b) < self.ATA_range):
            if self.adv_count >= 0:
                self.adv_count += 1
            else:
                self.adv_count = 1
        elif(dis < self.Dis_max) and (dis > self.Dis_min) \
                and (abs(self.AA_r) < self.AA_range) and (abs(self.ATA_r) < self.ATA_range):
            if self.adv_count <= 0:
                self.adv_count -= 1
            else:
                self.adv_count = -1
        else:
            self.adv_count = 0
        self.advs.append(self.adv_count)

        # reward shaping
        RA_b = 1 - ((1 - math.fabs(self.ATA_b) / 180) + (1 - math.fabs(self.AA_b) / 180))
        RA_r = 1 - ((1 - math.fabs(self.ATA_r) / 180) + (1 - math.fabs(self.AA_r) / 180))
        RD = math.exp(-(math.fabs(dis - ((self.Dis_max + self.Dis_min) / 2)) / 180 * 0.1))
        Rbl_b = RA_b * RD
        Rbl_r = RA_r * RD
        self.dis_count = Rbl_r
        self.old_fai_b = self.fai_b
        self.old_fai_r = self.fai_r
        self.fai_b = -0.01 * Rbl_b
        self.fai_r = -0.01 * Rbl_r

        # termination reward
        if self.adv_count >= 2:
            self.done = True
            self.success = 1
            self.reward_b = 2.0
            self.reward_r = -2.0
        elif self.adv_count <= -2:
            self.done = True
            self.success = -1
            self.reward_b = -2.0
            self.reward_r = 2.0
        elif (self.step_left <= 0):
            self.done = True
            self.success = 0
            self.reward_b = -1.0
            self.reward_r = -1.0
        elif (self.ac_pos_b[0] > self.AREA) or ((0 - self.ac_pos_b[0]) > self.AREA) or \
                (self.ac_pos_b[1] > self.AREA) or ((0 - self.ac_pos_b[1]) > self.AREA):
            self.done = True
            self.success = 0
            self.reward_b = -1.0
            self.reward_r = (self.fai_r - self.old_fai_r) - 0.001
        elif (self.ac_pos_r[0] > self.AREA) or ((0 - self.ac_pos_r[0]) > self.AREA) or \
                (self.ac_pos_r[1] > self.AREA) or ((0 - self.ac_pos_r[1]) > self.AREA):
            self.done = True
            self.success = 0
            self.reward_b = (self.fai_b - self.old_fai_b) - 0.001
            self.reward_r = -1.0
        else:
            self.done = False
            self.reward_b = (self.fai_b - self.old_fai_b) - 0.001
            self.reward_r = (self.fai_r - self.old_fai_r) - 0.001

        return s_b, s_r, self.reward_b, self.reward_r, self.done

    def creat_ALG(self):
        self.Tk = tk.Tk()
        self.Tk.title('1V1')
        self.Tk.canvas = tk.Canvas(self.Tk, bg='white',
                                height=self.SCOPE*2,
                                width=self.SCOPE*2)
        self.Tk.canvas.pack()

    def render(self):
        # red aircraft
        self.r_show = self.xyz2abc(self.ac_pos_r)
        self.r = self.Tk.canvas.create_oval(
            self.r_show[0] - 1, self.r_show[1] - 1,
            self.r_show[0] + 1, self.r_show[1] + 1,
            fill='red')

        # blue aircraft
        self.b_show = self.xyz2abc(self.ac_pos_b)
        self.b =  self.Tk.canvas.create_oval(
            self.b_show[0] - 1, self.b_show[1] - 1,
            self.b_show[0] + 1, self.b_show[1] + 1,
            fill='blue')

        self.Tk.update()
        time.sleep(0.01)
        if self.done:
            time.sleep(0.01)
            self.Tk.destroy()

    def getAngle(self,pos_r, pos_b, heading_r, heading_b):
        theta_br = 180 * math.atan2((pos_r[1] - pos_b[1]), (pos_r[0] - pos_b[0])) / math.pi
        theta_rb = 180 * math.atan2((pos_b[1] - pos_r[1]), (pos_b[0] - pos_r[0])) / math.pi
        if theta_br < 0:
            theta_br = 360 + theta_br
        if theta_rb < 0:
            theta_rb = 360 + theta_rb
        ATA = heading_b - theta_br
        AA = 180 + heading_r - theta_rb
        if ATA > 180:
            ATA = 360 - ATA
        elif ATA < -180:
            ATA = 360 + ATA
        if AA > 180:
            AA = 360 - AA
        elif AA < -180:
            AA = 360 + AA
        return ATA, AA

    def xyz2abc(self,pos):
        pos_show = np.array([0,0])
        pos_show[0] = pos[0] * self.SCALE + self.SCOPE
        pos_show[1] = self.SCOPE - pos[1] * self.SCALE
        return pos_show


#Test
if __name__ == '__main__':
    env = AirCombatEnv()
    s = env.reset_selfPlay()
    env.creat_ALG()
    env.render()
    while True:
        if env.done:
            print('-------------------------------------------------')
            s = env.reset_selfPlay()
            env.creat_ALG()
        #a = np.random.randint(0,3)
        a_b = 2
        a_r = 2
        env.step_selfPlay(a_b,a_r)
        env.render()